import datetime
import os

import argparse
import fnmatch
import shutil
import webbrowser

from sarcharts.lib.chartjs import ChartJS
from sarcharts.lib.chartsconf import ChartsConf
from sarcharts.lib.sadf import Sadf
from sarcharts.lib import util


class SarCharts:
    args = ()
    cwd = os.getcwd()
    C = ChartsConf()

    def valid_date(self, date):
        try:
            return datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            raise argparse.ArgumentTypeError(f"not a valid date: {date!r}")

    def valid_path(self, path):
        if os.path.exists(path):
            return path
        else:
            raise argparse.ArgumentTypeError(f"not a valid path: {path!r}")

    def default_sarfiles(self):
        files = []
        for f in os.listdir(self.cwd):
            if fnmatch.fnmatch(f, 'sa??'):
                files.append(f"{self.cwd}/{str(f)}")
        return files

    def __init__(self):
        self.parser = argparse.ArgumentParser(
            description="SarCharts gets \"sysstat\" files from provided"
            + " `sarfilespaths` and generates dynamic HTML Charts."
            )
        self.parser.add_argument(
            '-d',
            '--debug',
            help='Set debug level. Default `W`.',
            default='W',
            choices=['D', 'I', 'W', 'E']
            )
        self.parser.add_argument(
            '-f',
            '--fromdate',
            help='Read metric starting on this date.',
            default='1970-01-01 00:00:00',
            type=self.valid_date
            )
        self.parser.add_argument(
            '-o',
            '--outputpath',
            help='Path to put output files. Default `./sarcharts`.',
            default='.'
            )
        self.parser.add_argument(
            '-t',
            '--todate',
            help='Discard metrics after this date.',
            default='2099-01-01 00:00:00',
            type=self.valid_date
            )
        self.parser.add_argument(
            'sarfilespaths',
            help='`sa` file/s to parse. Default: `./sa??`.',
            default=self.default_sarfiles(),
            type=self.valid_path,
            nargs='*'
            )
        self.args = self.parser.parse_args()

        # create required files on outputpath
        self.args.outputpath = self.args.outputpath + "/sarcharts"
        if os.path.exists(self.args.outputpath):
            util.debug(self.args.debug, 'E',
                       f"Output path '{self.args.outputpath}' already exists.")
            # shutil.rmtree(self.args.outputpath)
        os.makedirs(self.args.outputpath + "/sar")
        shutil.copytree(os.path.dirname(
                        os.path.realpath(__file__)) + "/html",
                        self.args.outputpath + "/html")

    def main(self):
        # import ipdb; ipdb.set_trace()
        if len(self.args.sarfilespaths) > 0:
            sarfiles = util.get_filelist(self.args.sarfilespaths)
            util.debug(self.args.debug, 'D', "sarfiles: " + str(sarfiles))
            chartinfo = Sadf().sar_to_chartjs(
                self.args.debug, sarfiles, self.args.outputpath + "/sar",
                self.C.charts, self.args.fromdate, self.args.todate
                )
            ChartJS().write_files(
                self.C.charts, self.C.colors, chartinfo, self.args.outputpath
                )
            util.debug(self.args.debug, '',
                       "Open SarCharts in default browser.")
            webbrowser.open(self.args.outputpath + "/cpu.html", 0, True)
        else:
            self.parser.print_help()
            util.debug(self.args.debug, 'E',
                       "No valid `sa` files on provided path.")



import datetime
import os

import jinja2

from sarcharts.lib.progressbar import ProgressBar


class ChartJS:

    def write_files(self, charts, colors, chartinfo, outputpath):
        pb = ProgressBar()
        pb.all_entries = len(charts) * len(charts)
        pb.start_time = datetime.datetime.now()
        pbi = 0
        for chart, details in charts.items():
            pbi += 1
            pb.print_bar(pbi, f"Write {chart}.")
            context = {
                "chart": chart,
                "details": details,
                "pages": charts.keys(),
                "colors": colors,
                "notavailable": chartinfo['notavailable'],
                "hostname": chartinfo['hostname'],
                "firstdate": chartinfo['firstdate'],
                "lastdate": chartinfo['lastdate']
                }
            parent = (
                os.path.dirname(os.path.realpath(__file__)) + "/../templates/"
                )
            environment = (
                jinja2.Environment(loader=jinja2.FileSystemLoader(parent))
                )
            template = environment.get_template("chart.html")
            outputfile = outputpath + f"/{chart}.html"
            with open(outputfile, mode="w", encoding="utf-8") as results:
                results.write(template.render(context))
        pb.finish(f"  Write Output to {outputpath}.")



#!/usr/bin/python3
import os
import sys

try:
    sys.path.insert(0, os.getcwd())
    from sarcharts import SarCharts
except KeyboardInterrupt:
    raise SystemExit()


def main():
    SarCharts().main()



class ChartsConf:
    charts = {
        "cpu": {
            "arg": "-P ALL",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": [
                '%steal',
                '%idle'
                ]
            },
        "disk": {
            "arg": "-d",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": [
                'tps',
                'dtps',
                'bread/s',
                'bwrtn/s',
                'bdscd/s'
                ]
            },
        "hugepages": {
            "arg": "-H",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "inodes": {
            "arg": "-v",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "io": {
            "arg": "-b",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "interrrupts": {
            "arg": "-I",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "load": {
            "arg": "-q LOAD",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": [
                'runq-sz',
                'plist-sz',
                'blocked'
                ]
            },
        "memory": {
            "arg": "-r ALL",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": [
                'kbmemfree',
                'kbavail',
                'kbmemused',
                'kbbuffers',
                'kbcached',
                'kbcommit',
                '%commit',
                'kbactive',
                'kbinact',
                'kbdirty',
                'kbanonpg',
                'kbslab',
                'kbkstack',
                'kbpgtbl',
                'kbvmused'
                ]
            },
        "mount": {
            "arg": "-F",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netdevice": {
            "arg": "-n DEV",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": [
                'rxpck/s',
                'txpck/s',
                'rxcmp/s',
                'txcmp/s',
                'rxmcst/s',
                '%ifutil'
                ]
            },
        "netdevicee": {
            "arg": "-n EDEV",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netfiberchannel": {
            "arg": "-n FC",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "neticmp": {
            "arg": "-n ICMP",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "neticmpe": {
            "arg": "-n EICMP",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netip": {
            "arg": "-n IP",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netipe": {
            "arg": "-n EIP",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netip6": {
            "arg": "-n IP6",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netip6e": {
            "arg": "-n EIP6",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netnfs": {
            "arg": "-n NFS",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netsock": {
            "arg": "-n SOCK",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netsock6": {
            "arg": "-n SOCK6",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "nettcp": {
            "arg": "-n TCP",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "nettcpe": {
            "arg": "-n ETCP",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netudp": {
            "arg": "-n UDP",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "netudp6": {
            "arg": "-n UDP6",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "paging": {
            "arg": "-B",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": [
                'fault/s',
                'majflt/s',
                'pgfree/s',
                'pgscank/s',
                'pgscand/s',
                'pgsteal/s',
                '%vmeff'
                ]
            },
        "powermanagement": {
            "arg": "-m ALL",
            "multiple": True,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "swap": {
            "arg": "-S",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": [
                'kbswpfree',
                'kbswpused',
                'kbswpcad',
                '%swpcad'
                ]
            },
        "tasks": {
            "arg": "-w",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
        "tty": {
            "arg": "-y",
            "multiple": False,
            "datasets": {},
            "labels": [],
            "hidden": []
            },
              }
    colors = ['255, 99, 132',
              '255, 159, 64',
              '255, 205, 86',
              '75, 192, 192',
              '54, 162, 235',
              '153, 102, 255',
              '201, 203, 207',
              '153, 24, 44',
              '54, 157, 72',
              '75, 89, 123',
              '255, 22, 237',
              '99, 215, 99',
              '199, 215, 29',
              '68, 15, 229',
              '88, 115, 67',
              '149, 245, 44']



import datetime
import os
import re

from sarcharts.lib.progressbar import ProgressBar
from sarcharts.lib import util


class Sadf:

    def sar_to_csv(self, inputfile, arg, debuglevel):
        command = f"sadf -d {inputfile} -- {arg}"
        [stdout, stderr] = util.exec_command(debuglevel, command)
        if stderr:
            if "Try to convert it to current format" in stderr:
                # tf = tempfile.NamedTemporaryFile(prefix="sarcharts")
                command = f"sadf -c {inputfile} > /tmp/sarcharts.tmp"
                [stdout, stderr] = util.exec_command(debuglevel, command)
                return self.sar_to_csv(
                    "/tmp/sarcharts.tmp", arg, debuglevel
                    )
            elif "Requested activities not available" in stderr:
                util.debug(debuglevel, 'I', stderr.strip())
            else:
                util.debug(debuglevel, 'W', command)
                util.debug(debuglevel, 'W', stderr.strip())
        else:
            out = []
            for line in stdout.split("\n"):
                if line != "":
                    out.append(line.split(";"))
            return out

    def merge_sarfiles(self, debuglevel, sarfiles, outputpath, charts, dfrom, dto):
        pb = ProgressBar()
        pb.all_entries = len(charts) * len(sarfiles)
        pb.start_time = datetime.datetime.now()
        pbi = 0
        for k, v in charts.items():
            content = []
            csvfile = f"{outputpath}/{k}.csv"
            out = False
            for inputfile in sarfiles:
                pbi += 1
                pb.print_bar(
                    pbi,
                    "Get data from " + inputfile.split("/")[-1] + " " + k)
                out = self.sar_to_csv(inputfile, v['arg'], debuglevel)
                if out:
                    headers = out.pop(0)
                    util.debug(debuglevel, 'D',
                               f"Merge {inputfile} to {csvfile}")
                    content = content + out
            if out:
                with open(csvfile, "w") as f:
                    f.write(';'.join(headers) + "\n")
                    content.sort(key=lambda x: x[2])
                    for line in content:
                        if (line[2] != "timestamp" and util.in_date_range(
                                debuglevel, dfrom, dto, line[2])):
                            f.write(';'.join(line) + "\n")
        pb.finish("  Get data.")

    def sar_to_chartjs(
            self, debuglevel, sarfiles, outputpath, charts, dfrom, dto):
        self.merge_sarfiles(
            debuglevel, sarfiles, outputpath, charts, dfrom, dto)
        chartinfo = {
            "notavailable": [],
            "hostname": '',
            "firstdate": '',
            "lastdate": ''
            }
        pb = ProgressBar()
        pb.all_entries = len(charts)
        pb.start_time = datetime.datetime.now()
        pbi = 0
        for k, v in charts.items():
            pbi += 1
            pb.print_bar(pbi, f"Set data for {k} Chart.")
            csvfile = f"{outputpath}/{k}.csv"
            if not os.path.exists(csvfile):
                chartinfo['notavailable'].append(k)
            else:
                with open(csvfile) as f:
                    # set the first data field
                    datastart = 4 if charts[k]['multiple'] else 3
                    # get headers from first line
                    line = f.readline().strip()
                    headers = line.split(";")[datastart:]
                    # get first stats date
                    pos = f.tell()
                    line = f.readline().split(";")
                    chartinfo['hostname'] = line[0]
                    chartinfo['firstdate'] = line[2]
                    # seek file to first stats line
                    f.seek(pos)
                    for line in f:
                        if "LINUX-RESTART" in line or re.match(r"^#", line):
                            continue
                        fields = line.strip().split(";")
                        # set fake item on non multiple charts
                        item = fields[3] if charts[k]['multiple'] else ""
                        # add date field to Chart labels
                        if fields[2] not in charts[k]['labels']:
                            charts[k]['labels'].append(fields[2])
                        if item not in charts[k]['datasets'].keys():
                            charts[k]['datasets'][item] = []
                            for h in headers:
                                charts[k]['datasets'][item].append({
                                    "label": h,
                                    "values": []
                                    })
                        for i in range(len(fields[datastart:])):
                            charts[k]['datasets'][
                                item][i]['values'].append({
                                    'x': fields[2],
                                    'y': fields[i+datastart]
                                    })
                    if line != "":
                        chartinfo['lastdate'] = fields[2]
        pb.finish("  Set Data.")
        return chartinfo



import datetime

from datetime import datetime as mytime


class ProgressBar:
    all_entries: int
    start_time: datetime
    last_printed_tenth_of_percentage: int

    def __init__(self) -> None:
        self.all_entries = 0
        self.last_printed_tenth_of_percentage = 0

    def set_number_of_file_lines(self, log_file_name: str):
        try:
            with open(log_file_name, 'r') as file:
                self.all_entries = max(i for i, line in enumerate(file)) + 1
        except Exception as file_exception:
            print(file_exception)

        self.start_time = mytime.now()
        return self.all_entries

    def set_number_of_entries(self, number: int):
        self.all_entries = number
        self.start_time = mytime.now()

    def print_bar(self, done_lines: int, msg):
        if self.all_entries == 0:
            return
        tenth_of_percentage = int(1000 * (done_lines / self.all_entries))
        if self.last_printed_tenth_of_percentage >= tenth_of_percentage:
            return
        half_percentage = int((tenth_of_percentage/1000) * (30 + 1))
        new_bar = chr(9608) * half_percentage + " " * (30 - half_percentage)
        # now = mytime.now()
        # left = (
        #    (self.all_entries - done_lines)
        #     * (now - self.start_time) / done_lines
        #    )
        # sec = int(left.total_seconds())
        # time_left = "Estimated time left: "
        # if sec > 60:
        #    text += f"{format(int(sec / 60))} min "
        # text += f"{format(int(sec % 79)+1)} sec       "
        print(" " * 79, end="\r\r")
        text = f"\r|{new_bar}| {tenth_of_percentage/10:.0f} %  " + msg
        print(text, end="\r\r")

        # print(tenth_of_percentage)
        if tenth_of_percentage == 999:
            print(" " * 79, end="\r\r")
        self.last_printed_tenth_of_percentage = tenth_of_percentage

    def finish(self, msg):
        print(" " * 79, end="\r\r")
        print(msg)



import datetime
import os
import re
import sys

import fnmatch
from pathlib import Path
import subprocess


def debug(debuglevel, sev, msg):
    C = {
        'I': '\033[0;34m',
        'D': '\033[01;36m',
        '': '\033[0;32m',
        'W': '\033[93m',
        'E': '\033[0;31m',
        'RESET': '\033[0m'
        }
    levels = {'D': 0,
              'I': 1,
              'W': 2,
              'E': 3,
              '': 4
              }
    if sev == "":
        print(f"  {str(msg)}")
    elif levels[sev] >= levels[debuglevel]:
        print(f"{C[sev]}[{sev}]{C['RESET']} {str(msg)}")
    if sev == 'E':
        sys.exit(1)


def exec_command(debuglevel, cmd):
    debug(debuglevel, "D", "execcommand: " + cmd)
    p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE)
    stdout, stderr = p.communicate()
    stdout = str(stdout.decode("utf-8"))
    stderr = str(stderr.decode("utf-8"))
    # if stderr != "":
    #    print(cmd + "\n" + stderr)
    #    sys.exit(1)
    return [stdout, stderr]


def get_filelist(filepaths):
    files = []
    for path in filepaths:
        # path is a folder
        if Path(path).is_dir():
            for f in os.listdir(path):
                if fnmatch.fnmatch(f, 'sa??'):
                    files.append(f"{path}/{str(f)}")
        # path is a file
        elif Path(path).is_file():
            files.append(str(path))
    return files


def sortfiles_by_mtime(files):
    details = {}
    for f in files:
        mtime = os.path.getmtime(f)
        details[mtime] = f
    # 'sorted' returns a 'set' hence no duplicates
    # but it should be converted to a list
    return list(dict(sorted(details.items())).values())


def valid_date(s: str) -> datetime.datetime:
    try:
        return datetime.datetime.strptime(s, "%Y-%m-%d %H:%M:%S")
    except ValueError:
        return False


def is_valid_date(debuglevel, d):
    if re.match(r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} UTC', d):
        return True
    else:
        debug(debuglevel, 'E', "ERROR: date '" + d
              + "' doesn't match %Y-%m-%d %H:%M:%S UTC")


def in_date_range(debuglevel, dfrom, dto, d):
    if is_valid_date(debuglevel, d):
        d = datetime.datetime.strptime(d, '%Y-%m-%d %H:%M:%S UTC')
        if dfrom and dto:
            if d >= dfrom and d <= dto:
                return True
        else:
            return True
